import WalletConnectProvider from "@walletconnect/web3-provider";
import Web3 from "web3";
import {
  AVAILABLE_CHAINS,
  CONTRACT_ADDRESS,
  IChainData,
} from "../configs/config_chain";
import {
  ACCOUNT_CHANGE,
  EventBus,
  NEED_CHANGE_CHAIN,
  NEED_LOGIN,
  NEED_NONCE,
  SHOW_CHAIN_MODAL,
} from "../utils/event-bus";
import { hasMetamask, toHexChainId } from "../utils/chain.util";
import { AllChains } from "../configs/allchain";
import { ERC20ABI } from "../configs/contracts";
import { TransactionReceipt } from "web3-core";
import { singleton } from "../decorator/singleton.decorator";
import { isMobile } from "../utils/browser";
import { UserModule } from "../module/UserModule";
import { getItem, removeItem, setItem } from "../utils/cookies";

const EIP721_DOMAIN_DATA = [
  { name: "name", type: "string" },
  { name: "version", type: "string" },
  { name: "chainId", type: "uint256" },
  { name: "verifyingContract", type: "address" },
];

const CACHE_KEY = "cebg_chain_cache_key";

@singleton
export class Blockchain {
  provider: any;
  web3: Web3;
  currentChain = 0;
  // 0: null, 1: metamask, 2: walletconnect
  walletType = 0;
  dataCached = false;
  instanceCacheMap: Map<string, any>;
  public chainMap: Map<number, IChainData> = new Map();
  public rpc: any = {};

  constructor() {
    const allChainMap: Map<number, IChainData> = new Map();
    for (const d of AllChains) {
      const id = d.id;
      this.rpc[id] = d.rpc;
      allChainMap.set(id, d);
    }
    for (const id of AVAILABLE_CHAINS) {
      this.chainMap.set(id, allChainMap.get(id)!);
    }
    this.loadCachedProvider();
    this.instanceCacheMap = new Map();
    // UserModule.updateChainID(chainId)
    EventBus.$on(NEED_LOGIN, this.connect.bind(this));
  }

  get isWalletConnect() {
    return !!this.walletType && !!this.currentChain;
  }

  get mallAddress() {
    return CONTRACT_ADDRESS[this.currentChain]?.mall;
  }

  public async updateAccount() {
    const accounts = await this.web3.eth.getAccounts();
    if (accounts && accounts.length > 0) {
      UserModule.updateAccount(accounts[0]);
    }
  }

  public get hexChainId() {
    return toHexChainId(this.currentChain);
  }

  public async chainSelected(id: number) {
    if (!this.chainMap.has(id)) {
      return;
    }
    this.currentChain = id;
    if (this.provider) {
      await this.switchEthereumChain();
    } else {
      await this.connectWallet(true);
    }
  }

  public async connectWallet(isManual: boolean) {
    console.log(
      "begin connect to wallet: ",
      this.walletType,
      this.currentChain
    );
    if (this.walletType === 1) {
      this.provider = await this.connectMetaMask();
    } else if (this.walletType === 2) {
      this.provider = await this.connectWalletConnect();
    }
    if (!this.provider) {
      return;
    }
    this.web3 = new Web3(this.provider);
    const chainId = await this.web3.eth.getChainId();
    // await this.checkChain(chainId)

    this.subscribeToEvents();
    const accounts = await this.web3.eth.getAccounts();
    if (accounts && accounts.length > 0) {
      UserModule.updateAccount(accounts[0]);
    }
    if (!this.currentChain) this.currentChain = chainId;
    this.saveProvider();
    UserModule.updateChainID(chainId);
    UserModule.updateWalletStatus(true);
    console.log("current login chain: ", chainId);
    console.log("accountsLogin: ", accounts, UserModule.accountId);
    if (isManual) {
      EventBus.$emit(NEED_NONCE);
    } else {
      UserModule.updateStep(1);
    }
    return { account: accounts[0], chainId };
  }

  /**
   * check whether special chainId is supported by config
   * show chain picker when chainId is not supported and current wallet is meatmask
   * show message box when chainId is not supported and current wallet is wallet connect
   * @param {number} chainId
   * @return {Promise<void>}
   * @private
   */
  private async checkChain(chainId: number) {
    if (!this.chainMap.has(chainId)) {
      // if (this.walletType === 1) {
      try {
        await this.selectChain();
      } catch (err) {
        await this.disconnect();
      }
      // } else if (this.walletType === 2) {
      //   await this.disconnect()
      //   this.walletType = 2
      //   await MessageBox.alert(
      //     'You need to connect to supported network',
      //     'Wrong Network',
      //     {
      //       confirmButtonText: 'Confirm'
      //     })
      //   await this.connectWallet(true)
      // }
    }
  }

  public async connect(isManual = false) {
    if (isMobile()) {
      if (hasMetamask()) {
        this.walletType = 1;
      } else {
        this.walletType = 2;
      }
    } else {
      if (hasMetamask()) {
        if (isManual && !this.walletType) {
          this.walletType = await this.selectWallet();
          return;
        }
      } else {
        this.walletType = 2;
      }
    }
    if (isManual || this.isWalletConnect) {
      await this.connectWallet(isManual);
    }
  }

  /**
   * show wallet picker
   * @return {Promise<number>}
   * @private
   */
  private selectWallet(): Promise<number> {
    return new Promise((resolve, reject) => {
      EventBus.$emit(SHOW_CHAIN_MODAL, {
        confirm: (id: number) => {
          console.log("select wallet: ", id);
          resolve && resolve(id);
        },
        cancel: (reason: any) => {
          console.log("cancel select wallet: ", reason);
          reject && reject(reason);
        },
      });
    });
  }

  /**
   * show chain picker
   * @return {Promise<number>}
   * @private
   */
  private selectChain(): Promise<number> {
    return new Promise((resolve, reject) => {
      EventBus.$emit(NEED_CHANGE_CHAIN, {
        confirm: async (id: number) => {
          console.log("select chain: ", id);
          this.currentChain = id;
          if (this.provider) {
            await this.switchEthereumChain();
          }
          resolve && resolve(id);
        },
        cancel: (reason: any) => {
          console.log("cancel select chain: ", reason);
          reject && reject(reason);
        },
      });
    });
  }

  public async connectMetaMask() {
    let provider = null;
    if (typeof window.ethereum !== "undefined") {
      provider = window.ethereum;
      try {
        await provider.request({ method: "eth_requestAccounts" });
      } catch (error) {
        if (error.code === -32002) {
          // const hexChainId = toHexChainId(this.currentChain)
          // await provider.request({
          //   method: 'wallet_switchEthereumChain',
          //   params: [{ chainId: hexChainId }]
          // })
          throw new Error("MeatMask not login, Open MeatMask and login first");
        } else {
          throw new Error("User Rejected");
        }
      }
    } else if (window.web3) {
      provider = window.web3.currentProvider;
    } else if (window.celo) {
      provider = window.celo;
    } else {
      throw new Error("No Web3 Provider found");
    }
    return provider;
  }

  public async connectWalletConnect() {
    const provider = new WalletConnectProvider({
      infuraId: process.env.VUE_APP_WALLET_INFURAID,
      rpc: this.rpc,
    });
    try {
      await provider.enable();
    } catch (err) {
      console.log("connect to wallet connect error: ", err);
      await Promise.reject(err);
    }

    return provider;
  }

  public async getContractInstance(address: string, abi: any = ERC20ABI) {
    if (!this.instanceCacheMap.has(address)) {
      const instance = new this.web3.eth.Contract(abi, address, {
        from: UserModule.accountId,
      });
      this.instanceCacheMap.set(address, instance);
    }
    return this.instanceCacheMap.get(address);
  }

  public clearCachedProvider() {
    console.log("clear cached provider");
    removeItem(CACHE_KEY);
  }

  public loadCachedProvider() {
    const dataStr = getItem(CACHE_KEY);

    if (dataStr) {
      try {
        const data = JSON.parse(dataStr);
        this.walletType = data.walletType;
        // if (this.chainMap.has(data.chainId)) {
        this.currentChain = data.chainId;

        // }
      } catch (err) {
        console.log("err parse cached json");
        this.clearCachedProvider();
      }
    }
  }

  public saveProvider() {
    const data = {
      walletType: this.walletType,
      chainId: this.currentChain,
    };
    setItem(CACHE_KEY, JSON.stringify(data));
  }

  public async disconnect() {
    try {
      await UserModule.LogOut();
      await this.provider?.disconnect();
    } catch (err) {}
    this.currentChain = 0;
    this.walletType = 0;
    this.instanceCacheMap.clear();
    this.clearCachedProvider();
    UserModule.updateStep(0);
    UserModule.updateChainID(0);
    UserModule.updateNonce("");
    UserModule.updateAccount("");
    UserModule.updateWalletStatus(false);
  }

  public subscribeToEvents = () => {
    // Subscribe to accounts change
    this.provider.on("accountsChanged", async (accounts: string[]) => {
      console.log("accountsChanged: ", accounts);
      if (accounts && accounts.length > 0) {
        if (UserModule.accountId !== accounts[0]) {
          this.clearCachedProvider();
          await UserModule.LogOut();
          EventBus.$emit(ACCOUNT_CHANGE, accounts[0]);
          let event = new Event("account_change", {
            bubbles: true,
            cancelable: false,
          });
          window.dispatchEvent(event);
          setTimeout(() => {
            location.reload();
          }, 1000);
        }
      }
    });

    // Subscribe to chainId change
    this.provider.on("chainChanged", async (chainId: string) => {
      const chainIdNum = parseInt(chainId);
      console.log("chainChanged", chainId, chainIdNum);
      await this.checkChain(chainIdNum);
      this.currentChain = chainIdNum;
      this.saveProvider();
      UserModule.updateChainID(this.currentChain);
    });

    // Subscribe to session disconnection
    this.provider.on("disconnect", (err: any) => {
      console.log("disconnect", err);
    });
  };

  /**
   * change chain of metamask
   * @param {number} chainId
   * @param {() => void} cb
   * @return {Promise<void>}
   */
  async switchEthereumChain(chainId?: number, cb?: () => void) {
    chainId = chainId || this.currentChain;
    const hexChainId = toHexChainId(chainId);
    const onChainChange = (chainId: string) => {
      console.log("switchEthereumChain: ", chainId);
      this.provider.removeListener("chainChanged", onChainChange);
      cb && cb();
    };
    this.provider.on("chainChanged", onChainChange);
    try {
      await this.provider.request({
        method: "wallet_switchEthereumChain",
        params: [{ chainId: hexChainId }],
      });
      console.log("switch chain success");
    } catch (e) {
      console.log("switch chain error: ", e);
      if (e.code === 4902 || e.message.indexOf("Unrecognized chain ID") >= 0) {
        try {
          const data = this.chainMap.get(chainId)!;
          await this.provider.request({
            method: "wallet_addEthereumChain",
            params: [
              {
                chainId: hexChainId,
                chainName: data.name,
                nativeCurrency: {
                  name: data.symbol,
                  symbol: data.symbol,
                  decimals: data.decimals || 18,
                },
                blockExplorerUrls: [data.explorerurl],
                rpcUrls: [data.rpc],
              },
            ],
          });
          console.log("add chain success");
        } catch (addError) {
          console.error("add chain error: ", addError);
          this.provider.removeListener("chainChanged", onChainChange);
        }
      }
      // console.error(e)
    }
  }

  public async getTransactionReceipt(txHash: string) {
    return this.web3.eth.getTransactionReceipt(txHash);
  }

  public async getTxConfirms(txhash: string) {
    const receipt: TransactionReceipt = await this.getTransactionReceipt(
      txhash
    );
    const latest = await this.web3.eth.getBlockNumber();
    return latest - receipt.blockNumber + 1;
  }

  public async signPresale({
    type,
    paymentTokenAddress,
    price,
    buyerAddress,
    netId,
  }: {
    type: number | string;
    paymentTokenAddress: string;
    price: any;
    buyerAddress: string;
    netId?: any;
  }) {
    const nonce = (Math.random() * 100000) | 0;

    const signMsg = {
      item: type,
      token: paymentTokenAddress,
      price: price,
      salt: nonce,
    };
    netId = netId || (await this.web3.eth.getChainId());
    const signObj = {
      types: {
        EIP712Domain: EIP721_DOMAIN_DATA,
        set: [
          { name: "item", type: "uint256" },
          { name: "token", type: "address" },
          { name: "price", type: "uint256" },
          { name: "salt", type: "uint256" },
        ],
      },
      primaryType: "set",
      domain: {
        name: "BEBoxMall",
        version: "1",
        chainId: netId,
        verifyingContract: this.mallAddress,
      },
      message: signMsg,
    };
    const signature = await this.signData(signObj, buyerAddress);
    return { nonce, signature };
  }

  public async increaseAllowance(address: string, price: string) {
    const coinInstance: any = await this.getContractInstance(address);
    const res = await coinInstance.methods
      .increaseAllowance(this.mallAddress, price)
      .send({ gas: 1000000 });
    console.log("increaseAllowance: ", res);
  }

  /**
   * @param {string} address
   * @param {string | null} account
   * @return {Promise<any>}
   */
  public async getBalance(address: string, account: string | null) {
    account = account || UserModule.accountId;
    const coinInstance: any = await this.getContractInstance(address);
    return await coinInstance.methods.balanceOf(account).call();
  }

  public async transferToAccount(
    account: string,
    amount: number,
    address: string
  ) {
    const amountBN = this.web3.utils.toBN(this.web3.utils.toWei(amount + ""));
    const coinInstance: any = await this.getContractInstance(address);
    return coinInstance.methods
      .transfer(account, amountBN)
      .send({ gas: 1000000 });
  }

  public async signData(signObj: any, signer: string) {
    const msgParams = JSON.stringify(signObj);
    const from = signer;
    console.log("clicked, sending personal sign req", "from", from, msgParams);
    const params = [from, msgParams];
    const result: any = await this.sendCmd({
      method: "eth_signTypedData_v4",
      params,
      from,
    });
    console.log(result);
    return result.result;
  }

  public async sendCmd({ method, params, from }: any) {
    return new Promise((resolve, reject) => {
      // eslint-disable-next-line @typescript-eslint/ban-ts-comment
      // @ts-ignore
      this.web3.currentProvider.sendAsync(
        {
          method,
          params,
          from,
        },
        async function (err: any, result: any) {
          if (err) {
            reject && reject(err);
            return;
          }
          resolve && resolve(result);
        }
      );
    });
  }
}
